/*
 *   Copyright (C) 2009,2010 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "OpenG2ProtocolHandler.h"
#include "DStarRepeaterConfig.h"
#include "URIUSBController.h"
#include "SerialController.h"
#include "ProtocolHandler.h"
#include "K8055Controller.h"
#include "DummyController.h"
#include "DStarRepeaterD.h"
#include "DStarDefines.h"
#include "Version.h"
#include "Logger.h"

#include <wx/cmdline.h>
#include <wx/wfstream.h>
#include <wx/fileconf.h>
#include <wx/filename.h>

const wxChar* NAME_PARAM       = wxT("Repeater Name");
const wxChar* NOLOGGING_SWITCH = wxT("nolog");

int main(int argc, char** argv)
{
	bool res = ::wxInitialize();
	if (!res) {
		::fprintf(stderr, "dstarrepeaterd: failed to initialise the wxWidgets library, exiting\n");
		return -1;
	}

	wxCmdLineParser parser(argc, argv);
	parser.AddSwitch(NOLOGGING_SWITCH, wxEmptyString, wxEmptyString, wxCMD_LINE_PARAM_OPTIONAL);
	parser.AddParam(NAME_PARAM, wxCMD_LINE_VAL_STRING, wxCMD_LINE_PARAM_OPTIONAL);

	int cmd = parser.Parse();
	if (cmd != 0) {
		::wxUninitialize();
		return 0;
	}

	bool nolog = parser.Found(NOLOGGING_SWITCH);

	wxString name;
	if (parser.GetParamCount() > 0U)
		name = parser.GetParam(0U);

	CDStarRepeaterD repeater(nolog, name);

	if (!repeater.init()) {
		::wxUninitialize();
		return 1;
	}

	repeater.run();

	::wxUninitialize();
	return 0;
}

CDStarRepeaterD::CDStarRepeaterD(bool nolog, const wxString& name) :
m_name(name),
m_nolog(nolog),
m_thread(NULL)
{
}

CDStarRepeaterD::~CDStarRepeaterD()
{
}

bool CDStarRepeaterD::init()
{
	if (!m_nolog) {
		wxString logBaseName = LOG_BASE_NAME;
		if (!m_name.IsEmpty()) {
			logBaseName.Append(wxT("_"));
			logBaseName.Append(m_name);
		}

		wxFileName loggerName(wxFileName::GetHomeDir(), logBaseName, wxT("log"));
		wxLog* log = new CLogger(loggerName.GetFullPath());
		wxLog::SetActiveTarget(log);
	} else {
		wxLog::SetActiveTarget(new wxLogStderr);
	}

	::wxLogInfo(wxT("Starting ") + APPLICATION_NAME + wxT(" - ") + VERSION);

	return createThread();
}

void CDStarRepeaterD::run()
{
	m_thread->Create();
	m_thread->Run();

	m_thread->Wait();
	delete m_thread;

	::wxLogInfo(APPLICATION_NAME + wxT(" is exiting"));
}

bool CDStarRepeaterD::createThread()
{
	wxFileName fileName(wxFileName::GetHomeDir(), CONFIG_FILE_NAME);
	wxFileInputStream stream(fileName.GetFullPath());
	if (!stream.IsOk()) {
		::wxLogError(wxT("Cannot open the config file - %s"), fileName.GetFullPath().c_str());
		return false;
	}

	CDStarRepeaterConfig config(new wxFileConfig(stream), m_name);

	m_thread = new CDStarRepeaterThread;

	wxString callsign, gateway;
	DSTAR_MODE mode;
	bool ack;
	config.getCallsign(callsign, gateway, mode, ack);
	m_thread->setCallsign(callsign, gateway, mode, ack);
	::wxLogInfo(wxT("Callsign set to \"%s\", gateway set to \"%s\", mode: %d, ack: %d"), callsign.c_str(), gateway.c_str(), int(mode), ack);

	wxString gatewayAddress, localAddress;
	unsigned int gatewayPort, localPort, networkDelay;
	config.getNetwork(gatewayAddress, gatewayPort, localAddress, localPort, networkDelay);
	::wxLogInfo(wxT("Gateway set to %s:%u, local set to %s:%u, network delay: %u ms"), gatewayAddress.c_str(), gatewayPort, localAddress.c_str(), localPort, networkDelay * 20U);

	if (!gatewayAddress.IsEmpty()) {
		IProtocolHandler* handler = new COpenG2ProtocolHandler(gatewayAddress, gatewayPort, localAddress, localPort);
		m_thread->setProtocolHandler(handler, networkDelay);
	}

	unsigned int timeout, ackTime;
	config.getTimes(timeout, ackTime);
	m_thread->setTimes(timeout, ackTime);
	::wxLogInfo(wxT("Timeout set to %u, Ack time set to %u"), timeout, ackTime);

	wxString readDevice, writeDevice;
	bool rxInvert, txInvert;
	wxFloat32 level;
	config.getRadio(readDevice, writeDevice, level, rxInvert, txInvert);
	::wxLogInfo(wxT("Soundcard set to %s:%s, level: %.1f, GMSK Inversion set to %d:%d"), readDevice.c_str(), writeDevice.c_str(), level, rxInvert, txInvert);

	if (!readDevice.IsEmpty() && !writeDevice.IsEmpty()) {
#if defined(__WINDOWS__)
		CSoundCardReaderWriter* soundcard = new CSoundCardReaderWriter(readDevice, writeDevice, DSTAR_RADIO_SAMPLE_RATE, DSTAR_RADIO_BLOCK_SIZE);
#else
		CSoundCardReaderWriter* soundcard = new CSoundCardReaderWriter(readDevice, writeDevice, DSTAR_RADIO_SAMPLE_RATE, 64U);
#endif
		soundcard->setCallback(m_thread, 0U);

		bool res = soundcard->open();
		if (!res) {
			::wxLogError(wxT("Cannot open the sound card"));
			return false;
		}

		m_thread->setSoundCard(soundcard, level, rxInvert, txInvert);
	}

	wxString type;
	int pttDelay;
	bool pttInvert;
	config.getController(type, pttDelay, pttInvert);
	::wxLogInfo(wxT("Controller set to %s, ptt delay: %d ms, PTT Inversion set to %d"), type.c_str(), pttDelay * 20, pttInvert);

	CExternalController* controller = NULL;

	wxString port;
	if (type.StartsWith(wxT("Velleman K8055 - "), &port)) {
		unsigned long num;
		port.ToULong(&num);
		controller = new CExternalController(new CK8055Controller(num), pttInvert, false);
	} else if (type.StartsWith(wxT("URI USB - "), &port)) {
		unsigned long num;
		port.ToULong(&num);
		controller = new CExternalController(new CURIUSBController(num), pttInvert, false);
	} else if (type.StartsWith(wxT("Serial - "), &port)) {
		controller = new CExternalController(new CSerialController(port), pttInvert, false);
	} else {
		controller = new CExternalController(new CDummyController, pttInvert, false);
	}

	bool res = controller->open();
	if (!res) {
		::wxLogError(wxT("Cannot open the hardware interface - %s"), type.c_str());
		return false;
	}

	m_thread->setController(controller, pttDelay);

	bool out1, out2, out3, out4;
	config.getOutputs(out1, out2, out3, out4);
	m_thread->setOutputs(out1, out2, out3, out4);
	::wxLogInfo(wxT("Output 1 = %d, output 2 = %d, output 3 = %d, output 4 = %d"), out1, out2, out3, out4);

	bool enabled;
	wxString rpt1Callsign, rpt2Callsign;
	wxString shutdown, startup;
	wxString command1, command1Line;
	wxString command2, command2Line;
	wxString output1, output2, output3, output4;
	config.getControl(enabled, rpt1Callsign, rpt2Callsign, shutdown, startup, command1, command1Line, command2, command2Line, output1, output2, output3, output4);
	m_thread->setControl(enabled, rpt1Callsign, rpt2Callsign, shutdown, startup, command1, command1Line, command2, command2Line, output1, output2, output3, output4);
	::wxLogInfo(wxT("Control: enabled: %d, RPT1: %s, RPT2: %s, shutdown: %s, startup: %s, command1: %s = %s, command2: %s = %s, output1: %s, output2: %s, output3: %s, output4: %s"), enabled, rpt1Callsign.c_str(), rpt2Callsign.c_str(), shutdown.c_str(), startup.c_str(), command1.c_str(), command1Line.c_str(), command2.c_str(), command2Line.c_str(), output1.c_str(), output2.c_str(), output3.c_str(), output4.c_str());

	unsigned int hangTime;
	config.getActiveHang(hangTime);
	m_thread->setActiveHang(hangTime);
	::wxLogInfo(wxT("Active Hang: time: %u"), hangTime);

	m_thread->Create();
	m_thread->SetPriority(100U);
	m_thread->Run();

	return true;
}
